#ifndef MATH3D_H

#include "pch.h"
#include "bipointer.h"


using namespace std;

//#define NAN_CHECK check_and_throw_nan()
#define NAN_CHECK {}

class VERTEX
{
private:
	//buffer for float[3] conversions
	float float3[3];

	void check_and_throw_nan() const;
	
public:
	float x;
	float y;
	float z;
	VERTEX operator+ (const VERTEX & v) const;
	VERTEX operator- (const VERTEX & v) const;
	bool operator!= (const VERTEX & v) const {return !equals(v);}
	friend ostream& operator <<(ostream &os,const VERTEX &v);
	VERTEX();
	//VERTEX(const VERTEX & other) {Set(other);NAN_CHECK;}
	//VERTEX(float nx, float ny, float nz) {Set(nx, ny, nz);NAN_CHECK;}
	VERTEX(const VERTEX & other) : x(other.x),y(other.y),z(other.z) {NAN_CHECK;}
	VERTEX(float nx, float ny, float nz) : x(nx),y(ny),z(nz) {NAN_CHECK;}
	VERTEX & operator=(const VERTEX & other) {Set(other);NAN_CHECK;return *this;}

	float len() const;
	VERTEX normalize() const;
	VERTEX cross(const VERTEX & b) const;
	inline void Scale(float scalar) {x = x*scalar;y = y*scalar;z = z*scalar;}
	VERTEX ScaleR(float scalar) const;
	float dot(const VERTEX & b) const;
	inline void Set(const VERTEX & other) {memcpy(this, &(other), sizeof(VERTEX) );NAN_CHECK;}
	void Set(float nx, float ny, float nz);
	void Set(const float * f3);
	VERTEX InvertR();
	float * v3();
	void zero();
	bool equals(const VERTEX & other) const;

	inline void flabs() {if (x < 0) x = -x; if (y < 0) y = -y; if (z < 0) z = -z;}

	inline float v(int i) {if (i == 2) return z; if (i == 1) return y; return x;}

	void DebugPrint() const {DebugPrint(cout);}
	void DebugPrint(ostream & out) const;

	VERTEX interpolatewith(VERTEX other, float percent);
	
	bool nan() const;
	bool IsZero() const {VERTEX z; return equals(z);}
};

class MATRIX3
{
public:
	void ProjectionMatrix(VERTEX normal);
	VERTEX Multiply(VERTEX vert);
//private:
	VERTEX row[3];
};

class QUATERNION;

class MATRIX4 : public BIPOINTER_TARGET
{
public:
	float data[16];

	MATRIX4() {LoadIdentity();}
	MATRIX4(const MATRIX4 & other) {Set(other);}
	const MATRIX4 & operator=(const MATRIX4 & other) {Set(other); return *this;}
	MATRIX4 Multiply(const MATRIX4 & other) const;
	void DebugPrint() const;
	void LoadIdentity();
	void Set(const float * newdata);
	void Set(const MATRIX4 & other) {Set(other.data);}
	void Set(const QUATERNION & quat);
	void Translate(const VERTEX & trans);
	bool operator==(const MATRIX4 & other) {return Equals(other);}
	bool Equals(const MATRIX4 & other);
	VERTEX TransformVertexIn(const VERTEX & v) const;
	VERTEX TransformVertexOut(const VERTEX & v) const;
	void Scale(float scalar);
};

class QUATERNION
{
private:
	bool cachematvalid;
	float cachemat[16];
	void check_and_throw_nan(); //const;
	
public:
	//the 4 vectors we need to represent the quaternion
	float w;
	float x;
	float y;
	float z;

	QUATERNION();			//the constructor loads an identity matrix
	QUATERNION(float neww, float newx, float newy, float newz) : w(neww),x(newx),y(newy),z(newz) {}
	void LoadMultIdent();	//load the [1,(0,0,0)] quaternion
	float Norm();			//return the magnitude of the quaternion
	void Normalize();		//normalize the quaternion
	void GetMat(float * destmat) const;	//convert to a matrix and store in (float) destmat[16]
	void GetMat(GLdouble * destmat) const;	//convert to a matrix and store in (gldouble) destmat[16]
	void GetAxisAngle(float * aangle);	//convert to axis angle form and store in (float) aangle[4] (a,x,y,z)
	void Multiply(QUATERNION quat2);	//multiply this quaternion by quat2
	void PostMultiply(QUATERNION quat2);	//multiply quat2 by this quaternion
	void SetAxisAngle(float a, float ax, float ay, float az);
	void Rotate(float a, float ax, float ay, float az);
	void PreRotate(float a, float ax, float ay, float az);
	void SetEuler(float a, float b, float c);
	QUATERNION ReturnConjugate() const;
	void LookAt(float eyeX, float eyeY, float eyeZ, float centerX, float centerY, 
		   float centerZ, float upX, float upY, float upZ);
	void SetMat(float * m1);
	VERTEX RotateVec(const VERTEX & vec) const;
	float GetAngleBetween(QUATERNION quat2);
	VERTEX GetEuler();
	//spherical linear interpolation between this quat and another quat
	//QUATERNION Slerp(QUATERNION quat2, float t);
	QUATERNION QuatSlerp(QUATERNION quat2, float t);
	QUATERNION Add(QUATERNION quat2);
	QUATERNION MultScalar(float scalar);
	void DebugPrint();
	VERTEX RelativeMove(VERTEX input);
	bool equals(const QUATERNION & other) const;

	QUATERNION operator*(const QUATERNION &quat2 );	//multiply override
	bool IsIdentity() const {QUATERNION ident; return equals(ident);}
	bool nan() const;
};

//axis-aligned bounding box
class AABB
{
private:
	VERTEX pos;
	VERTEX center;
	VERTEX size;

	inline float fabs(const float num) const {if (num >= 0) return num; else return -num;}

public:
	AABB(const AABB & other) : pos(other.pos),center(other.center),size(other.size) {}
	AABB() {}
	void SetTo(const AABB & other) {pos = other.pos;center=other.center;size=other.size;}
	void SetFromCorners(const VERTEX & c1, const VERTEX & c2);
	bool IntersectRay(const VERTEX & orig, const VERTEX & d) const;
	bool IntersectSegment(const VERTEX & orig, const VERTEX & d, float seglen) const;
	bool IntersectAABB(const AABB & other) const;
	void CombineWith(const AABB & other);

	const VERTEX & GetPos() const {return pos;}
	const VERTEX & GetSize() const {return size;}
	const VERTEX & GetCenter() const {return center;}
	
	void DebugPrint();
	
	void GetVerts(VERTEX * eightverts);
};

#define MATH3D_H
#endif
