#include "autox_track.h"
#include "game.h"
extern GAME game;


AUTOX_TRACK::~AUTOX_TRACK()
{
        std::vector<VERTEX*>::iterator i;
        for (i=cones.begin();i<cones.end();i++) delete *i;
}

void AUTOX_TRACK::Load(string trackname, bool reverse)
{
	CONFIGFILE trackconfig;
	trackconfig.Load((game.settings.GetDataDir() + "/tracks/" + trackname + "/track.txt").c_str());

        float tvert[3];
	trackconfig.GetParam("origin", tvert);
	field_origin.Set(tvert);
	trackconfig.GetParam("width", field_width);
	trackconfig.GetParam("height", field_height);

	CalcStart();

	string strcullfaces;
	trackconfig.GetParam("cull faces", strcullfaces);
	
	if (strcullfaces == "no")
		cullfaces = false;
	else
		cullfaces = true;
	
	string strhoriz;
	trackconfig.GetParam("vertical tracking skyboxes", strhoriz);
	if (strhoriz == "no")
		verticaltrackingskyboxes = false;
	else
		verticaltrackingskyboxes = true;
	game.objects.SetVerticalSkyboxTracking(verticaltrackingskyboxes);
	
	ClearRoads();
	CreateRoad();
	
	lapsequence.clear();
	//first lapmarker is always the 1st patch of the 1st (and only) road
	lapsequence.push_back(GetBezier(0, 0));
	//next lapmarker is somewhere half way of the 1st (and only) road
	lapsequence.push_back(GetBezier(0, cones.size()/4));
	
	if (reverse)
	{
		Reverse();
		direction = false;
	}
	else direction = true;
	
	float f;
	trackconfig.GetParam("non-treaded friction coefficient", f);
	friction1 = f;
	trackconfig.GetParam("treaded friction coefficient", f);
	friction2 = f;

	// calculate distance from starting line for each patch to account for those tracks
	// where starting line is not on the 1st patch of the road
	// note this only update the road with lap sequence 0 on it
	BEZIER* start_patch = GetLapSequence(0);
	if (start_patch)
	{
		start_patch->dist_from_start = 0.0;
		BEZIER* curr_patch = start_patch->next_patch;
		float total_dist = start_patch->length;
		int count = 0;
		while ( curr_patch && curr_patch != start_patch)
		{
			count++;
			curr_patch->dist_from_start = total_dist;
			total_dist += curr_patch->length;
			curr_patch = curr_patch->next_patch;
		}
	}
}

VERTEX AUTOX_TRACK::Map2Field(VERTEX* pos)
{
	VERTEX field_pos;
	field_pos.x = field_origin.x + pos->x * field_width;
	field_pos.y = field_origin.y + pos->y;
	field_pos.z = field_origin.z + pos->z * field_height;
	return field_pos;
}

void AUTOX_TRACK::CalcStart()
{
	if (cones.empty())
	{
		startloc.clear();
		VERTEX sl(0.5,0,0.5);
		SetStart(Map2Field(&sl));
		QUATERNION so;
		SetStartOrientation(so);
	}
	else
	{
		VERTEX start_cone_l = Map2Field(cones[0]);
		VERTEX start_cone_r = Map2Field(cones[1]);
	
		startloc.clear();
		VERTEX sl = (start_cone_l + start_cone_r).ScaleR(0.5);
		//sl.y += 0.5;
		SetStart(sl);
	
		VERTEX next_cone_l = Map2Field(cones[2]);
		VERTEX next_cone_r = Map2Field(cones[3]);
	
		VERTEX nl = (next_cone_l + next_cone_r).ScaleR(0.5);
	
		startquat.clear();
		QUATERNION so;
		so.LookAt(sl.x, sl.y, sl.z, nl.x, nl.y, nl.z, 0, 1, 0);
		SetStartOrientation(so);
	}
}	

void AUTOX_TRACK::CreateRoad()
{
	ROADSTRIP * newroad = AddNewRoad();

	BEZIER * newpatch = NULL;
        VERTEX corner[4];
        int index = 0;
        std::vector<VERTEX*>::iterator i;
        for (i=cones.begin();i<cones.end();i++)
        {
		corner[index] = Map2Field(*i);
		index++;
		if (index == 4)
		{
			BEZIER temppatch;
			temppatch.SetFromCorners(corner[2], corner[3], corner[0], corner[1]);
			newpatch = newroad->Add(temppatch);
			corner[0] = corner[2];
			corner[1] = corner[3];
			index = 2;
		}
	}

	//load a default big square if the user didn't place any cones
	if (cones.empty())
	{
		float offset = 0;
		
		VERTEX temp;
		temp.Set(offset,0,offset);
		corner[0] = Map2Field(&temp);
		temp.Set(1-offset,0,offset);
		corner[1] = Map2Field(&temp);
		temp.Set(offset,0,1-offset);
		corner[2] = Map2Field(&temp);
		temp.Set(1-offset,0,1-offset);
		corner[3] = Map2Field(&temp);
		
		BEZIER temppatch;
		temppatch.SetFromCorners(corner[2], corner[3], corner[0], corner[1]);
		newpatch = newroad->Add(temppatch);
		corner[0] = corner[2];
		corner[1] = corner[3];
		index = 2;
	}
	
	newpatch->next_patch = NULL;

        //close the track
	if (!open_track && (cones.size()%2 == 0) && (cones.size() > 1) && !cones.empty())
	{
		corner[2] = Map2Field(cones[0]);
		corner[3] = Map2Field(cones[1]);
		BEZIER temppatch;
		temppatch.SetFromCorners(corner[2], corner[3], corner[0], corner[1]);
		newpatch = newroad->Add(temppatch);
		newpatch->Attach(*(newroad->GetFirstPatch()));
	}
	
	newroad->GenerateCollisionData();
}

void AUTOX_TRACK::CreateCones()
{
	string m = "Roadcone.joe";
	string t = "Roadcone.png";
	bool mip = true;
	bool fb = false;
	bool sb = false;
	bool blend = false;
	float bl = 0.0;
	float bm = 0.0;
	VERTEX p;
	p.zero();
	float r = 0.0;
	bool d = false;
	bool c = false;
	float f1 = 0.7;
	float f2 = 1.0;
	float rr = 5.0;	
	float rd = 20.0;

	JOEPACK pack;
	JOEPACK * packptr = NULL;
	string objectpath = game.settings.GetDataDir() + "/tracks/parkinglot/objects";

	if (pack.LoadPack(objectpath+"/objects.jpk"))
	{
		packptr = &pack;
	}

	SCENENODE *parentnode = const_cast<SCENENODE*>(&game.objects.GetSceneNode());

        std::vector<VERTEX*>::iterator i;
        for (i=cones.begin();i<cones.end();i++)
        {
		SCENENODE &conenode = parentnode->AddNode();
		VERTEX cone_pos = conenode.TransformIntoLocalSpace(Map2Field(*i));
		conenode.GetTransform().SetTranslation(cone_pos);

		game.objects.Add(p, r, m, t, mip, fb, sb, d, c, blend, packptr, f1, f2, bl, bm, rr, rd, &conenode);
	}

	if (packptr != NULL) packptr->ClosePack();

}
