//  Vamos Automotive Simulator
//  Copyright (C) 2001--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <vamos/world/World.h>
#include "singleton.h"
#include "common.h"
#include "game.h"
extern GAME game;

using Vamos_Geometry::Three_Vector;
using Vamos_Geometry::Three_Matrix;

//-----------------------------------------------------------------------------
Vamos_World::
Times::Times () :
		m_current ( 0.0 ),
		m_previous ( 0.0 ),
		m_best ( 0.0 ),
		m_difference ( 0.0 )
{}

void Vamos_World::
Times::update ( double time_step )
{
	m_current += time_step;
}

void Vamos_World::
Times::finalize ()
{
	if ( m_best != 0.0 )
	{
		m_difference = m_current - m_best;
	}

	if ( ( m_current < m_best ) || ( m_best == 0.0 ) )
	{
		m_best = m_current;
	}

	m_previous = m_current;
	m_current = 0.0;
}

void Vamos_World::
Times::reset ()
{
	m_current = 0.0;
}

//-----------------------------------------------------------------------------
Vamos_World::
Timing_Info::Timing_Info () :
		m_sector ( -1 ),
		m_previous_sector ( -1 ),
		m_distance ( 0.0 )
{}

void Vamos_World::
Timing_Info::update_sector_info ( int sector )
{
	if ( sector == int ( ma_sector_times.size () ) )
	{ // We're entering a sector for the first time.
		ma_sector_times.push_back ( Times () );
	}

	if ( ( sector != m_sector ) && ( sector != -1 ) )
	{ // We have left a sector.
		if ( m_sector != -1 )
		{
			ma_sector_times [m_sector].finalize ();

			if ( sector == 0 )
			{ // We have finished a lap.
				m_lap_times.finalize ();
			}
		}
		m_previous_sector = m_sector;
		m_sector = sector;
	}
}

void Vamos_World::
Timing_Info::update_times ( int sector, double time_step )
{
	m_lap_times.update ( time_step );
	if ( sector != -1 )
	{
		ma_sector_times [sector].update ( time_step );
	}
}

void Vamos_World::
Timing_Info::update ( double time_step, double distance, int sector )
{
	update_sector_info ( sector );
	update_times ( sector, time_step );
	m_distance = distance;
}

void Vamos_World::
Timing_Info::reset ()
{
	m_lap_times.reset ();
	ma_sector_times [m_sector].reset ();

	m_sector = -1;
	m_previous_sector = -1;
	m_distance = 0.0;
}

//-----------------------------------------------------------------------------
void Vamos_World::
Car_Information::reset ()
{
	timing.reset ();
	segment_index = 0;
	car->reset ();
	dist_from_start = 0.0;
	lap_number = 0;
	race_position = 0;
	end_race = false;
	drifting = false;
	temp_drift_score = 0.0;
	temp_bonus_score = 0.0;
	max_drift_angle = 0.0;
	max_drift_speed = 0.0;
	drift_score = 0;
}

//-----------------------------------------------------------------------------
Vamos_World::
World::World ( TRACK * track )
		:
		m_focused_car_index ( 0 ),
		p_track ( track ),
		m_gravity ( -9.8 )
{}

extern bool verbose_output;

Vamos_World::
World::~World ()
{
	for ( std::vector <Car_Information>::iterator it = m_cars.begin ();
	        it != m_cars.end ();
	        it++ )
	{
		delete it->car;
	}
	//std::cout << (void*)mp_track << std::endl;
	//delete mp_track;
}

extern float FrameTime();

bool BoxContactSort ( const PHYSICSCONTACT & lhs, const PHYSICSCONTACT & rhs )
{
	const VERTEX lnorm = lhs.GetContactNormal();
	const VERTEX rnorm = rhs.GetContactNormal();

	float ldepth = lhs.GetContactDepth();
	float rdepth = rhs.GetContactDepth();

	float l = ldepth * sqrt ( lnorm.x*lnorm.x+lnorm.z*lnorm.z );
	float r = rdepth * sqrt ( rnorm.x*rnorm.x+rnorm.z*rnorm.z );

	return l < r;

	//return lnorm.y > rnorm.y;
}

void Vamos_World::
World::interact ( Vamos_Body::Car* car, size_t segment_index )
{
	Vamos_Geometry::Material * tmatroad = new Vamos_Geometry::Material (
	                                          Vamos_Geometry::Material::ASPHALT, 1.0, 1.0,
	                                          1.0, 0.0,
	                                          0.0, 0.0,
	                                          0 );
	Vamos_Geometry::Material * tmatwall = new Vamos_Geometry::Material (
	                                          Vamos_Geometry::Material::CONCRETE, 1.0, 0.5,
	                                          1.0, 0.0,
	                                          0.0, 0.0,
	                                          0 );
	Vamos_Geometry::Material_Handle material = tmatroad;
	Vamos_Geometry::Material_Handle materialwall = tmatwall;

	size_t i;

	const int FRAMESKIP_SPEEDUP ( game.GetCollisionFrameskip());

	PHYSICS_CONTACT_CACHE_INFO & last_contact = last_contact_map[car];

	//do additional collisions for walls and stuff
	bool verbose = false;
	int loopcount = 0;
	bool col = true;
	const int maxloops = 10;
	while ( col && loopcount < maxloops && last_contact.framecount == 0 )
	{
		loopcount++;
		col = false;

		//get the collision mesh
		JOEMODEL * mod = car->GetCollisionModel();
		if ( mod != NULL )
		{
			VERTEX col_center;
			col_center.Set ( mod->GetBBOX().GetCenter().x, mod->GetBBOX().GetCenter().z, -mod->GetBBOX().GetCenter().y );
			VERTEX col_size;
			col_size.Set ( mod->GetBBOX().GetSize().y, mod->GetBBOX().GetSize().x, mod->GetBBOX().GetSize().z );

			double angle;
			Vamos_Geometry::Three_Vector aa = car->chassis().axis_angle ( &angle );
			QUATERNION carorientation, boxorientation;
			//carorientation.SetAxisAngle(angle*(3.141593/180), -aa[0], -aa[2], aa[1]);
			//carorientation.Rotate(3.141593/2.0,0,1,0);
			//carorientation.Rotate(-3.141593/2.0, 1,0,0);

			carorientation.SetAxisAngle ( angle*0.01745329, aa[0], aa[1], aa[2] );
			carorientation.Rotate ( -3.141593/2.0, 1,0,0 );
			//carorientation.Rotate(3.141593/2.0,0,1,0);

			boxorientation = carorientation;

			//Vamos_Geometry::Three_Vector cp = car->chassis().position();
			VERTEX carposition;
			//carposition.Set(cp[0], cp[2], -cp[1]);
			if ( !car->GetOriginNode() )
				throw EXCEPTION ( __FILE__, __LINE__, "Car has no scene node" );
			carposition = car->GetOriginNode()->TransformIntoWorldSpace();

			PHYSICSCOLLISIONSETTINGS settings;
			settings.SetMapCollide ( true );
			list <PHYSICSCONTACT> contacts;
			VERTEX boxpos = carposition + col_center;
			//col_size.Set(2.5,0.7,2.5);
			//boxpos.y += 0.5;
			game.physics.CollideBox ( boxpos, boxorientation, col_size, contacts, settings );

			if ( !contacts.empty() )
			{
				col = true;

				if ( verbose )
				{
					cout << "Loop " << loopcount << endl;
					cout << "Collisions: " << contacts.size() << endl;
					cout << "BBOX Center: ";col_center.DebugPrint();
					cout << "BBOX Size: ";col_size.DebugPrint();
					cout << "Camera: "; game.cam.position.ScaleR ( -1.0 ).DebugPrint();
					cout << "Car Position: ";carposition.DebugPrint();
					cout << "BBOX Position: ";boxpos.DebugPrint();
				}

				//sort by contact depth
				contacts.sort();

				//sort for the optimal contact
				//contacts.sort(BoxContactSort);

				//after sort, the contact at the back is the one we want
				PHYSICSCONTACT & curcol = contacts.back();
				//cout << "Minimum depth: " << contacts.front().GetContactDepth() << endl;

				if ( verbose )
				{
					cout << "Collision position: ";curcol.GetContactPosition().DebugPrint();
					cout << "Collision normal: ";curcol.GetContactNormal().DebugPrint();
					cout << "Collision depth: " << curcol.GetContactDepth() << endl;
				}

				VERTEX colpt ( curcol.GetContactPosition() );
				VERTEX normal ( curcol.GetContactNormal() );
				float depth ( curcol.GetContactDepth() );

				//depth += 0.1;
				//depth += 0.001;
				depth += 0.0001;

				if ( ( colpt - boxpos ).dot ( normal ) )
				{
					//cout << "Flipping Normal" << endl;
					normal.Scale ( -1.0 );
				}
				else throw EXCEPTION ( __FILE__, __LINE__, "Unusual box collision normal behavior" );

				Three_Vector colpos;
				Three_Vector norm;

				colpos.m_vec[0] = colpt.x;
				colpos.m_vec[1] = -colpt.z;
				colpos.m_vec[2] = colpt.y;

				norm.m_vec[0] = normal.x;
				norm.m_vec[1] = -normal.z;
				norm.m_vec[2] = normal.y;

				norm = -norm; //norm is flipped for some reason

				car->chassis().single_point_contact ( colpos, depth, norm, materialwall, game.FrameTime() *FRAMESKIP_SPEEDUP );
				car->Update();
			}
		}
	}

	//keep the wheels on the track
	i = 0;
	int wheelcount = 0;
	for ( i = 0; i < 4; i++ )
		car->SetColPatch ( i, NULL );
	for ( std::vector <Vamos_Body::Particle*>::iterator
	        it = car->chassis ().particles ().begin ();
	        it != car->chassis ().particles ().end ();
	        it++, i++ )
	{
		if ( ! ( *it )->single_contact () )
		{
			const Three_Vector& pos = car->chassis ().contact_position ( *it );
			//const Three_Vector& lastpos = car->chassis ().last_contact_position (*it);
			//double bump_parameter = car->distance_traveled () + (*it)->position () [0];

			VERTEX particlepos;
			particlepos.Set ( pos.m_vec[0],pos.m_vec[2],-pos.m_vec[1] );

			//cout << i << ". ";particlepos.DebugPrint();

			bool con = false;
			Three_Vector norm;
			double depth = 0;
			VERTEX normal ( 0,1,0 );
			VERTEX origin = particlepos;
			/*cout << "Origin: "; origin.DebugPrint();
			cout << "Camera: "; game.cam.position.ScaleR(-1.0).DebugPrint();*/
			origin.y += 1.0;
			BEZIER * tcolpatch = NULL;
			OBJECTNODE * tcolnode = NULL;

			/*depth = 2.0-particlepos.y;
			car->SetColPatch(wheelcount, tcolpatch);
			car->SetColParams(wheelcount, 1.0, 1.0, 1.0, 0.0);
			wheelcount++;
			if (depth > 0)
			{
				norm.m_vec[0] = normal.x;
				norm.m_vec[1] = -normal.z;
				norm.m_vec[2] = normal.y;
				car->chassis ().contact (*it, depth, norm, material);
				//car->Update();
			}*/

			if ( last_contact.framecount == 0 )
			{
				depth = p_track->ElevationSeg ( origin, normal, 10.0, tcolpatch, tcolnode ) - particlepos.y;

				if ( depth > 0 )
				{
					last_contact.wheel_had_contact[wheelcount] = true;

					VERTEX dir ( 0,-1,0 );
					VERTEX tempnorm = normal;
					if ( dir.dot ( normal ) > 0 )
						tempnorm.Scale ( -1.0 );

					last_contact.last_wheel_contact_position[wheelcount] = origin;
					last_contact.last_wheel_contact_position[wheelcount].y = depth + particlepos.y;
					last_contact.last_wheel_contact_normal[wheelcount] = tempnorm;

					last_contact.last_wheel_tcolpatch[wheelcount] = tcolpatch;
					last_contact.last_wheel_tcolnode[wheelcount] = tcolnode;
				}
				else
				{
					last_contact.wheel_had_contact[wheelcount] = false;
				}
			}
			else
			{
				if ( last_contact.wheel_had_contact[wheelcount] )
				{
					//simple approximation
					//float newelev = last_wheel_contact_position[wheelcount].y;

					//enhanced plane-based approximation
					float D = - last_contact.last_wheel_contact_normal[wheelcount].dot ( last_contact.last_wheel_contact_position[wheelcount] );
					float Pn_dot_Rd = last_contact.last_wheel_contact_normal[wheelcount].dot ( VERTEX ( 0,-1,0 ) );
					bool newcon ( true );
					if ( Pn_dot_Rd != 0 )
					{
						float t = - ( last_contact.last_wheel_contact_normal[wheelcount].dot ( origin ) + D ) / ( Pn_dot_Rd );
						if ( t >= 0 )
						{
							float newelev = origin.y + -1*t;

							depth = newelev - particlepos.y;

							tcolpatch = last_contact.last_wheel_tcolpatch[wheelcount];
							tcolnode = last_contact.last_wheel_tcolnode[wheelcount];
						}
						else
							newcon = false;
					}
					else
						newcon = false;

					if ( !newcon )
					{
						//cout << "extrapolated plane doesn't collide: " << Pn_dot_Rd << "," << D << endl;
						depth = 0;
						tcolpatch = NULL;
						tcolnode = NULL;
					}
				}
				else
				{
					depth = 0;
					tcolpatch = NULL;
					tcolnode = NULL;
				}
			}

			//cout << framecount << "," << wheelcount << "," << depth << endl;

			//cout << "WORLD::INTERACT Initial depth=" << depth << endl;

			car->SetColPatch ( wheelcount, tcolpatch );
			//if (wheelcount == 3) cout << particlepos.y + depth << endl;

			//calculate friction coefficients from collision data and track data
			float f1 = 1.0;
			float f2 = 1.0;
			float rr = 1.0;
			float rd = 0.0;
			if ( tcolpatch != NULL )
			{
				p_track->GetFrictionParams ( f1, f2 );
			}
			else if ( tcolnode != NULL )
			{
				f1 = tcolnode->friction1;
				f2 = tcolnode->friction2;
				//cout << "Setting friction wheel " << wheelcount << " to " << f1 << "," << f2 << endl;

				//add in bump effects
				float posx = origin.x;
				float posz = origin.z;
				float phase = 2*3.141593* ( posx+posz ) / tcolnode->bumplength;
				float shift = 2.0*sin ( phase*1.414214 );
				float amplitude = 0.25*tcolnode->bumpmag;
				float bumpoffset = amplitude * ( sin ( phase + shift ) + sin ( 1.414214*phase ) - 2.0 );
				depth += bumpoffset;

				rr = tcolnode->rolling_resistance_factor;
				rd = tcolnode->rolling_drag;
			}
			car->SetColParams ( wheelcount, f1, f2, rr, rd );

			wheelcount++;

			//depth = p_track->Elevation(particlepos, normal) - particlepos.y;

			if ( depth < 0 || depth >= 0 )
			{
				//depth = -particlepos.y;
				//normal.Set(0,1,0);

				VERTEX dir ( 0,-1,0 );
				if ( dir.dot ( normal ) > 0 )
				{
					normal.Scale ( -1.0 );
					//throw EXCEPTION(__FILE__, __LINE__, "Unusual elevation collision normal behavior");
					//cout << "Unusual elevation collision normal behavior" << endl;
				}

				if ( depth > 0 )
					con = true;
				norm.m_vec[0] = normal.x;
				norm.m_vec[1] = -normal.z;
				norm.m_vec[2] = normal.y;

				//normal.DebugPrint();

				if ( con )
				{
					if ( depth > 1.0 )
					{
						cout << "Large suspension displacement detected: " << depth << endl;
					}

					//cout << p_track->Elevation(particlepos, normal) << endl;
					//cout << "WORLD::INTERACT collision depth: " << depth << ", " << particlepos.y << endl;
					//particlepos.DebugPrint();	cout << endl;
					//Vamos_Geometry::Material_Handle mat = &info.material;
					car->chassis ().contact ( *it, depth,
					                          norm, material );
					car->Update();
					//car->chassis().propagate_contact();
				}
			}
			else
				cout << "Not contacting with NaN surface" << endl;
		}
	}

	last_contact.framecount++;

	if ( last_contact.framecount >= FRAMESKIP_SPEEDUP )
		last_contact.framecount = 0;

	// Handle air resistance.
	i = 0;
	for ( std::vector <Vamos_Body::Particle*>::iterator
	        it = car->chassis ().particles ().begin ();
	        it != car->chassis ().particles ().end ();
	        it++, i++ )
	{
		car->chassis ().wind ( *it, Three_Vector ( 0,0,0 )
		                       - car->chassis ().velocity ( *it ),
		                       1.2 );
	}

}

void Vamos_World::
World::collide ( Vamos_Body::Car* car1, Vamos_Body::Car* car2 )
{
	// Handle collisions between the contact points of car1 and the
	// crash box of car 2.
	for ( std::vector <Vamos_Body::Particle*>::iterator
	        it = car1->chassis ().particles ().begin ();
	        it != car1->chassis ().particles ().end ();
	        it++ )
	{
		if ( car2->collision ( car1->chassis ().transform_out ( ( *it )->position () ) ) )
		{
			std::cout << "boom" << std::endl;
		}
	}
}
void Vamos_World::
World::reset ()
{
	reset ( false );
}

// Place the car back on the track at its current position.
void Vamos_World::
World::reset ( bool all )
{
	//if (all)
	{
		unsigned int i = 0;
		for ( std::vector <Vamos_World::Car_Information>::iterator carit = m_cars.begin ();
		        carit != m_cars.end ();
		        carit++ )
		{
			reset ( & ( *carit ), i );
			i++;
		}
	}
}

void Vamos_World::
World::reset ( Car_Information * cartoreset, unsigned int index )
{
//	size_t& segment_index = cartoreset->segment_index;

	// Find the position and orientation specified by the track.
//	const Three_Vector& car_pos = cartoreset->car->chassis ().position ();
	/*const Three_Vector& position =
	mp_track->reset_position (car_pos, segment_index);
	const Three_Matrix& orientation = 
	mp_track->reset_orientation (car_pos, segment_index);*/

	VERTEX vpos = p_track->GetStart ( index );
	Three_Vector position ( vpos.x, -vpos.z, vpos.y );

	QUATERNION qstart;
	/*qstart = p_track->GetStartOrientation();
	//qstart = qstart.ReturnConjugate();
	//qstart.Rotate(3.141593/2.0, 0,0,1);
	//qstart.Rotate(3.141593/2.0, 0,1,0);
	//qstart.Rotate(-3.141593/2.0, 1,0,0);
	//qstart.Rotate(3.141593/2.0, 0,1,0);
	qstart.Rotate(-3.141593/2.0, 1,0,0);
	QUATERNION goofyfoot;
	goofyfoot.Rotate(3.141593/2.0, 1,0,0);
	qstart.Multiply(goofyfoot);*/

	double o3x3[3][3];
	float orient[16];

	qstart = p_track->GetStartOrientation ( index );
	QUATERNION qtemp = qstart;
	qstart.x = qtemp.x;
	qstart.y = qtemp.z;
	qstart.z = -qtemp.y;

	qstart.Rotate ( -3.141593/2.0, 0,0,1 );

	qstart.GetMat ( orient );
	bool flip = false;
	for ( int i = 0; i < 3; i++ )
	{
		if ( flip )
			o3x3[i][0] = orient[i+0];
		else
			o3x3[0][i] = orient[i+0];
	}
	for ( int i = 0; i < 3; i++ )
	{
		if ( flip )
			o3x3[i][1] = orient[i+4];
		else
			o3x3[1][i] = orient[i+4];
	}
	for ( int i = 0; i < 3; i++ )
	{
		if ( flip )
			o3x3[i][2] = orient[i+8];
		else
			o3x3[2][i] = orient[i+8];
	}
	//Three_Matrix orientation(1.0);
	//orientation.identity();
	Three_Matrix orientation ( o3x3 );

	cartoreset->car->reset ( position, orientation );

	// Adjust the vertical position so that at least one point on the
	// car is in contact with the road, and the rest are above it.

	//cartoreset->car->Update();

	//Three_Vector mpos =
	//cartoreset->car->chassis ().contact_position (*it);
	//double mdist = cartoreset->car->distance_traveled ()
	//+ (*it)->position () [0];
	//double elevation = mp_track->elevation (pos, dist, segment_index);

	VERTEX cp;
	//cp.Set(mpos.m_vec[0], mpos.m_vec[2], -mpos.m_vec[1]);
	double elevation;
	//elevation = p_track->Elevation (cp);

	//double max_diff = elevation - mpos [2];
	bool have_max = false;
	double max_diff = 0.0;

	//const Three_Vector& car_pos = cartoreset->car->chassis ().position ();

	for ( std::vector <Vamos_Body::Particle*>::iterator it =
	            cartoreset->car->chassis ().particles ().begin ();
	        it != cartoreset->car->chassis ().particles ().end ();
	        it++ )
	{
		Three_Vector pos = cartoreset->car->chassis ().contact_position ( *it );

		/*const Three_Vector & pos = (*it)->position();
		cp.Set(pos.m_vec[0],pos.m_vec[1],pos.m_vec[2]);
		cp = cartoreset->car->GetOriginNode()->TransformIntoWorldSpace(cp);*/

		cp.Set ( pos.m_vec[0], pos.m_vec[2], -pos.m_vec[1] );
		elevation = p_track->Elevation ( cp );

		double diff = elevation - cp.y;
		//double diff = pos[2];
		if ( diff > max_diff || !have_max )
		{
			max_diff = diff;
			have_max = true;
		}

		//cout << diff << endl;
	}

	// Set the car's position and orientation.
	cartoreset->car->reset ( position + Three_Vector ( 0.0, 0.0, max_diff+0.0 ), orientation );
	//cartoreset->car->reset (position, orientation);
	//cout << "Max diff: " << max_diff << endl;

	cartoreset->car->Update();
}

// Place the car back on the track at the starting line.
void Vamos_World::
World::restart ()
{
	focused_car ()->reset ();
}

// Set the acceleration due to gravity.  Always downward, regardless
// of sign.
void Vamos_World::
World::gravity ( double g )
{
	m_gravity = -std::abs ( g );
	if ( focused_car () != 0 )
	{
		focused_car ()->car->chassis ().
		gravity ( Three_Vector ( 0.0, 0.0, m_gravity ) );
	}
}

void Vamos_World::
World::add_car ( Vamos_Body::Car* car )
{
	car->chassis ().gravity ( Three_Vector ( 0.0, 0.0, m_gravity ) );
	m_cars.push_back ( Car_Information ( car ) );
}

void Vamos_World::
World::set_focused_car ( size_t index )
{
	assert ( index < m_cars.size () );
	m_focused_car_index = index;
}

void Vamos_World::
World::focus_other_car ( int delta )
{
	set_focused_car ( ( m_focused_car_index + delta ) % m_cars.size () );
}

Vamos_World::Car_Information* Vamos_World::
World::focused_car ()
{
	if ( m_focused_car_index >= m_cars.size () ) return 0;
	return &m_cars [m_focused_car_index];
}

void Vamos_World::World::update_race_position ()
{
	//bool reversetrack = true;

	for ( std::vector <Vamos_World::Car_Information>::iterator carit = m_cars.begin ();
	        carit != m_cars.end ();
	        carit++ )
	{
		BEZIER *curr_patch = carit->car->GetColPatch ( 0 );
		if ( !curr_patch ) curr_patch = carit->car->GetColPatch ( 1 ); //let's try the other wheel

		if ( curr_patch ) //only update if car is on track
		{
			Vamos_Geometry::Three_Vector pos = carit->car->chassis().position();
			Vamos_Geometry::Three_Vector back_left, back_right;

			if ( p_track->GetDirection() ) // true means forward direction
			{
				Vamos_Geometry::Three_Vector b_left ( curr_patch->points[3][0].x,
				                                      -curr_patch->points[3][0].z,
				                                      curr_patch->points[3][0].y );
				Vamos_Geometry::Three_Vector b_right ( curr_patch->points[3][3].x,
				                                       -curr_patch->points[3][3].z,
				                                       curr_patch->points[3][3].y );
				back_left = b_left;
				back_right = b_right;
			}
			else // false means reverse direction
			{
				Vamos_Geometry::Three_Vector b_left ( curr_patch->points[0][0].x,
				                                      -curr_patch->points[0][0].z,
				                                      curr_patch->points[0][0].y );
				Vamos_Geometry::Three_Vector b_right ( curr_patch->points[0][3].x,
				                                       -curr_patch->points[0][3].z,
				                                       curr_patch->points[0][3].y );
				back_left = b_left;
				back_right = b_right;
			}

			float dist_from_back = ( back_left - back_right ).perp_distance ( back_left, pos );

			//cout << curr_patch->dist_from_start << "," << dist_from_back << endl;

			carit->dist_from_start = curr_patch->dist_from_start + dist_from_back;
		}
	}

	//only update player car race position for now
	int player_car_lap = 0;
	float player_car_dist = 0;
	int index=0;

	for ( std::vector <Vamos_World::Car_Information>::iterator carit = m_cars.begin ();
	        carit != m_cars.end ();
	        carit++, index++ )
	{
		if ( index == 0 )
		{
			if ( carit->end_race ) return; //stop updating player position at end of race
			player_car_lap = carit->lap_number;
			player_car_dist = carit->dist_from_start;
			carit->race_position = 1;
		}
		else
		{
			if ( ( carit->lap_number > player_car_lap ) ||
			        ( ( carit->lap_number == player_car_lap ) &&
			          ( carit->dist_from_start > player_car_dist ) ) )
				m_cars.begin()->race_position++;
		}
	}
}

void Vamos_World::World::update_drift_score ( float update_time )
{
	for ( std::vector <Vamos_World::Car_Information>::iterator carit = m_cars.begin ();
	        carit != m_cars.end ();
	        carit++ )
	{
		bool is_drifting = false;
		bool spin_out = false;

		//make sure the car is not off track
		int wheel_count = 0;
		for ( int i=0;i<4;i++ )
		{
			if ( carit->car->GetColPatch ( i ) ) wheel_count++;
		}

		bool on_track = ( wheel_count > 1 );

		//car's direction on the horizontal plane
		Vamos_Geometry::Three_Vector car_orientation = carit->car->chassis().orientation().unit ( 0 );
		car_orientation[2] = 0;

		//car's velocity on the horizontal plane
		Vamos_Geometry::Three_Vector car_velocity = carit->car->chassis().cm_velocity();
		car_velocity[2] = 0;
		float car_speed = car_velocity.magnitude();

		//angle between car's direction and velocity
		double car_angle = car_orientation.angle ( car_velocity );

		if ( on_track )
		{
			//velocity must be above 10 m/s
			if ( car_speed > 10 )
			{
				//drift starts when the angle > 0.2 (around 11.5 degrees)
				//drift ends when the angle < 0.1 (aournd 5.7 degrees)
				float angle_threshold;
				if ( carit->drifting ) angle_threshold = 0.1;
				else angle_threshold = 0.2;

				is_drifting = ( car_angle > angle_threshold && car_angle < M_PI/2.0 );
				spin_out = ( car_angle > M_PI/2.0 );
			}
		}

		//calculate score
		if ( is_drifting )
		{
			//base score is the drift distance
			carit->temp_drift_score += update_time * car_speed;

			//bonus score calculation
			if ( car_speed > carit->max_drift_speed ) carit->max_drift_speed = car_speed;
			if ( car_angle > carit->max_drift_angle ) carit->max_drift_angle = car_angle;
			carit->temp_bonus_score = carit->temp_drift_score +
			                          carit->max_drift_speed / 2.0 +
			                          carit->max_drift_angle * 40.0 / M_PI;
		}
		else
		{
			if ( carit->drifting ) //accumulate score
			{
				if ( on_track && !spin_out && carit->temp_drift_score > 5.0 )
					carit->drift_score += ( unsigned int ) ( carit->temp_drift_score + carit->temp_bonus_score );
				carit->temp_drift_score = carit->temp_bonus_score = 0.0;
				carit->max_drift_speed = carit->max_drift_angle = 0.0;
			}
		}

		carit->drifting = is_drifting;
	}
}
